@@ -25,42 +25,60 @@ import math
 from copy import copy
 from decimal import Decimal
 from itertools import product
-from typing import TYPE_CHECKING, Optional, List, Tuple, Union, Any, Iterator
+from typing import TYPE_CHECKING, cast, Dict, Optional, List, Tuple, Union, \
+    Any, Iterator, SupportsFloat, Type
 import urllib.parse
-from xml.etree.ElementTree import Element
 
-from .exceptions import ElementPathError, ElementPathValueError, XPATH_ERROR_CODES
+from .exceptions import ElementPathError, ElementPathValueError, ElementPathNameError, \
+    ElementPathTypeError, ElementPathSyntaxError, MissingContextError, XPATH_ERROR_CODES
 from .helpers import ordinal
-from .namespaces import XQT_ERRORS_NAMESPACE, XSD_NAMESPACE, \
+from .namespaces import XQT_ERRORS_NAMESPACE, XSD_NAMESPACE, XSD_SCHEMA, \
     XPATH_FUNCTIONS_NAMESPACE, XPATH_MATH_FUNCTIONS_NAMESPACE, \
     XSD_ANY_TYPE, XSD_ANY_SIMPLE_TYPE, XSD_ANY_ATOMIC_TYPE, XSI_NIL
-from .xpath_nodes import XPathNode, ElementNode, TypedElement, AttributeNode, TextNode, \
+from .xpath_nodes import XPathNode, TypedElement, AttributeNode, TextNode, \
     NamespaceNode, TypedAttribute, is_etree_element, etree_iter_strings, \
     is_comment_node, is_processing_instruction_node, is_element_node, \
     is_document_node, is_xpath_node, is_schema_node
 from .datatypes import xsd10_atomic_types, xsd11_atomic_types, AbstractDateTime, \
     AnyURI, UntypedAtomic, Timezone, DateTime10, Date10, DayTimeDuration, Duration, \
-    Integer, DoubleProxy10, DoubleProxy, QName
+    Integer, DoubleProxy10, DoubleProxy, QName, DatetimeValueType, AtomicValueType, \
+    AnyAtomicType
+from .protocols import ElementProtocol, DocumentProtocol, \
+    XsdAttributeProtocol, XsdTypeProtocol, XMLSchemaProtocol
 from .schema_proxy import AbstractSchemaProxy
 from .tdop import Token, MultiLabel
 from .xpath_context import XPathContext, XPathSchemaContext
 
 if TYPE_CHECKING:
     from .xpath1 import XPath1Parser
+    from .xpath2 import XPath2Parser
+    from .xpath30 import XPath30Parser
+
+    XPathParserType = Union[XPath1Parser, XPath2Parser, XPath30Parser]
+else:
+    XPathParserType = Any
 
 UNICODE_CODEPOINT_COLLATION = "http://www.w3.org/2005/xpath-functions/collation/codepoint"
 XSD_SPECIAL_TYPES = {XSD_ANY_TYPE, XSD_ANY_SIMPLE_TYPE, XSD_ANY_ATOMIC_TYPE}
 
 # Type annotations aliases
 NargsType = Optional[Union[int, Tuple[int, Optional[int]]]]
+ClassCheckType = Union[Type[Any], Tuple[Type[Any], ...]]
+PrincipalNodeType = Union[ElementProtocol, AttributeNode, TypedAttribute, TypedElement]
+OperandsType = Tuple[Optional[AtomicValueType], Optional[AtomicValueType]]
+SelectResultType = Union[AtomicValueType, ElementProtocol, XsdAttributeProtocol, Tuple[str, str]]
+
+XPathTokenType = Union['XPathToken', 'XPathAxis', 'XPathFunction', 'XPathConstructor']
 
 
-class XPathToken(Token):
+class XPathToken(Token[XPathTokenType]):
     """Base class for XPath tokens."""
-    xsd_types = None  # for XPath 2.0+ schema types labeling
-    namespace = None  # for namespace binding of names and wildcards
+    parser: XPathParserType
+    xsd_types: Optional[Dict[str, Union[XsdTypeProtocol, List[XsdTypeProtocol]]]]
+    namespace: Optional[str]
 
-    parser: 'XPath1Parser'
+    xsd_types = None  # for XPath 2.0+ XML Schema types labeling
+    namespace = None  # for namespace binding of names and wildcards
 
     def evaluate(self, context: Optional[XPathContext] = None) -> Any:
         """
@@ -136,18 +154,18 @@ class XPathToken(Token):
             return False
         elif self.symbol != ':':
             return True
-        return not self[1].label.endswith('function')
+        return not self._items[1].label.endswith('function')
 
     ###
     # Tokens tree analysis methods
-    def iter_leaf_elements(self) -> Iterator[ElementNode]:
+    def iter_leaf_elements(self) -> Iterator[str]:
         """
         Iterates through the leaf elements of the token tree if there are any,
         returning QNames in prefixed format. A leaf element is an element
         positioned at last path step. Does not consider kind tests and wildcards.
         """
         if self.symbol in {'(name)', ':'}:
-            yield self.value
+            yield cast(str, self.value)
         elif self.symbol in ('//', '/'):
             if self._items[-1].symbol in {
                 '(name)', '*', ':', '..', '.', '[', 'self', 'child',
@@ -165,8 +183,13 @@ class XPathToken(Token):
 
     ###
     # Dynamic context methods
-    def get_argument(self, context, index=0, required=False, default_to_context=False,
-                     default=None, cls=None, promote=None):
+    def get_argument(self, context: Optional[XPathContext],
+                     index: int = 0,
+                     required: bool = False,
+                     default_to_context: bool = False,
+                     default: Optional[AtomicValueType] = None,
+                     cls: Optional[Type[Any]] = None,
+                     promote: Optional[ClassCheckType] = None) -> Any:
         """
         Get the argument value of a function of constructor token. A zero length sequence is
         converted to a `None` value. If the function has no argument returns the context's
@@ -182,6 +205,8 @@ class XPathToken(Token):
         :param cls: if a type is provided performs a type checking on item.
         :param promote: a class or a tuple of classes that are promoted to `cls` class.
         """
+        item: Union[None, ElementProtocol, DocumentProtocol, XPathNode, AnyAtomicType]
+
         try:
             selector = self._items[index].select
         except IndexError:
@@ -250,7 +275,8 @@ class XPathToken(Token):
 
         return item
 
-    def select_data_values(self, context=None):
+    def select_data_values(self, context: Optional[XPathContext] = None) \
+            -> Iterator[Optional[AtomicValueType]]:
         """
         Yields data value of selected items.
 
@@ -259,7 +285,8 @@ class XPathToken(Token):
         for item in self.select(context):
             yield self.data_value(item)
 
-    def atomization(self, context=None):
+    def atomization(self, context: Optional[XPathContext] = None) \
+            -> Iterator[AtomicValueType]:
         """
         Helper method for value atomization of a sequence.
 
@@ -275,7 +302,8 @@ class XPathToken(Token):
             else:
                 yield value
 
-    def get_atomized_operand(self, context=None):
+    def get_atomized_operand(self, context: Optional[XPathContext] = None) \
+            -> Optional[AtomicValueType]:
         """
         Get the atomized value for an XPath operator.
 
@@ -286,7 +314,7 @@ class XPathToken(Token):
         try:
             value = next(selector)
         except StopIteration:
-            return
+            return None
         else:
             item = getattr(context, 'item', None)
 
@@ -316,7 +344,7 @@ class XPathToken(Token):
                 msg = "atomized operand is a sequence of length greater than one"
                 raise self.wrong_context_type(msg)
 
-    def iter_comparison_data(self, context):
+    def iter_comparison_data(self, context: XPathContext) -> Iterator[OperandsType]:
         """
         Generates comparison data couples for the general comparison of sequences.
         Different sequences maybe generated with an XPath 2.0 parser, depending on
@@ -345,8 +373,10 @@ class XPathToken(Token):
 
             # Converts to float for lesser-greater operators (3.)
             if self.symbol in ('<', '<=', '>', '>='):
-                yield from product(map(float, map(self.data_value, operand1)),
-                                   map(float, map(self.data_value, operand2)))
+                yield from product(
+                    map(float, map(self.data_value, operand1)),  # type: ignore[arg-type]
+                    map(float, map(self.data_value, operand2)),  # type: ignore[arg-type]
+                )
                 return
             elif self.parser.version == '1.0':
                 yield from product(map(self.data_value, operand1), map(self.data_value, operand2))
@@ -364,7 +394,7 @@ class XPathToken(Token):
                 raise TypeError(msg.format(type(values[0]), type(values[1])))
             yield values
 
-    def select_results(self, context: Optional[XPathContext]) -> Iterator[Any]:
+    def select_results(self, context: Optional[XPathContext]) -> Iterator[SelectResultType]:
         """
         Generates formatted XPath results.
 
@@ -391,7 +421,8 @@ class XPathToken(Token):
                 else:
                     yield result.uri
 
-    def get_results(self, context):
+    def get_results(self, context: XPathContext) \
+            -> Union[List[Any], AtomicValueType]:
         """
         Returns formatted XPath results.
 
@@ -407,13 +438,14 @@ class XPathToken(Token):
             elif is_etree_element(res) or is_document_node(res) or is_schema_node(res):
                 return results
             elif self.label in ('function', 'literal'):
-                return res
+                return cast(AtomicValueType, res)
             else:
                 return results
         else:
             return results
 
-    def get_operands(self, context, cls=None):
+    def get_operands(self, context: XPathContext, cls: Optional[Type[Any]] = None) \
+            -> OperandsType:
         """
         Returns the operands for a binary operator. Float arguments are converted
         to decimal if the other argument is a `Decimal` instance.
@@ -427,13 +459,13 @@ class XPathToken(Token):
         if op1 is None:
             return None, None
         elif is_element_node(op1):
-            op1 = self[0].data_value(op1)
+            op1 = self._items[0].data_value(op1)
 
         op2 = self.get_argument(context, index=1, cls=cls)
         if op2 is None:
             return None, None
         elif is_element_node(op2):
-            op2 = self[1].data_value(op2)
+            op2 = self._items[1].data_value(op2)
 
         if isinstance(op1, AbstractDateTime) and isinstance(op2, AbstractDateTime):
             if context is not None and context.timezone is not None:
@@ -464,7 +496,9 @@ class XPathToken(Token):
 
         return op1, op2
 
-    def get_absolute_uri(self, uri, base_uri=None, as_string=True):
+    def get_absolute_uri(self, uri: str,
+                         base_uri: Optional[str] = None,
+                         as_string: bool = True) -> Union[str, AnyURI]:
         """
         Obtains an absolute URI from the argument and the static context.
 
@@ -480,6 +514,7 @@ class XPathToken(Token):
         if not base_uri:
             base_uri = self.parser.base_uri
 
+        url_parts: Union[urllib.parse.ParseResult, urllib.parse.SplitResult]
         url_parts = urllib.parse.urlparse(uri)
         if url_parts.scheme or url_parts.netloc \
                 or url_parts.path.startswith('/') \
@@ -495,7 +530,7 @@ class XPathToken(Token):
             return urllib.parse.urljoin(base_uri, uri)
         return AnyURI(urllib.parse.urljoin(base_uri, uri))
 
-    def get_namespace(self, prefix):
+    def get_namespace(self, prefix: str) -> str:
         """
         Resolves a prefix to a namespace raising an error (FONS0004) if the
         prefix is not found in the namespace map.
@@ -506,7 +541,7 @@ class XPathToken(Token):
             msg = 'no namespace found for prefix %r' % str(err)
             raise self.error('FONS0004', msg) from None
 
-    def bind_namespace(self, namespace):
+    def bind_namespace(self, namespace: str) -> None:
         """
         Bind a token with a namespace. The token has to be a name, a name wildcard,
         a function or a constructor, otherwise a syntax error is raised. Functions
@@ -537,7 +572,8 @@ class XPathToken(Token):
 
         self.namespace = namespace
 
-    def adjust_datetime(self, context, cls):
+    def adjust_datetime(self, context: XPathContext, cls: Type[DatetimeValueType]) \
+            -> Optional[Union[DatetimeValueType, DayTimeDuration]]:
         """
         XSD datetime adjust function helper.
 
@@ -547,11 +583,13 @@ class XPathToken(Token):
         or the adjusted XSD datetime instance.
         """
         timezone: Optional[Any]
+        item: Optional[DatetimeValueType]
+        _item: Union[DatetimeValueType, DayTimeDuration]
 
         if len(self) == 1:
             item = self.get_argument(context, cls=cls)
             if item is None:
-                return
+                return None
             timezone = getattr(context, 'timezone', None)
         else:
             item = self.get_argument(context, cls=cls)
@@ -563,25 +601,28 @@ class XPathToken(Token):
                 except ValueError as err:
                     raise self.error('FODT0003', str(err)) from None
             if item is None:
-                return
+                return None
 
+        _item = item
+        _tzinfo = _item.tzinfo
         try:
-            if item.tzinfo is not None and timezone is not None:
-                if isinstance(item, DateTime10):
-                    item += timezone.offset
+            if _tzinfo is not None and timezone is not None:
+                if isinstance(_item, DateTime10):
+                    _item += timezone.offset
                 elif not isinstance(item, Date10):
-                    item += timezone.offset - item.tzinfo.offset
-                elif timezone.offset < item.tzinfo.offset:
-                    item -= timezone.offset - item.tzinfo.offset
-                    item -= DayTimeDuration.fromstring('P1D')
+                    _item += timezone.offset - _tzinfo.offset
+                elif timezone.offset < _tzinfo.offset:
+                    _item -= timezone.offset - _tzinfo.offset
+                    _item -= DayTimeDuration.fromstring('P1D')
         except OverflowError as err:
             raise self.error('FODT0001', str(err)) from None
 
-        item.tzinfo = timezone
-        return item
+        if not isinstance(_item, DayTimeDuration):
+            _item.tzinfo = timezone
+        return _item
 
     @contextlib.contextmanager
-    def use_locale(self, collation):
+    def use_locale(self, collation: str) -> Iterator[None]:
         """A context manager for use a locale setting for string comparison in a code block."""
         loc = locale.getlocale(locale.LC_COLLATE)
         if collation == UNICODE_CODEPOINT_COLLATION:
@@ -600,7 +641,8 @@ class XPathToken(Token):
 
     ###
     # XSD types related methods
-    def select_xsd_nodes(self, schema_context, name):
+    def select_xsd_nodes(self, schema_context: XPathSchemaContext, name: str) \
+            -> Iterator[Union[None, TypedElement, TypedAttribute, XMLSchemaProtocol]]:
         """
         Selector for XSD nodes (elements, attributes and schemas). If there is
         a match with an attribute or an element the node's type is added to
@@ -612,25 +654,39 @@ class XPathToken(Token):
         :param schema_context: an XPathSchemaContext instance.
         :param name: a QName in extended format.
         """
+        xsd_node: Any
         for xsd_node in schema_context.iter_children_or_self():
             if xsd_node is None:
-                if name == schema_context.root.tag == '{%s}schema' % XSD_NAMESPACE:
+                if name == XSD_SCHEMA == schema_context.root.tag:
                     yield None
                 continue  # pragma: no cover
 
             try:
                 if isinstance(xsd_node, AttributeNode):
-                    if xsd_node.value.is_matching(name):
+                    if isinstance(xsd_node.value, str):
+                        if xsd_node.name != name:
+                            continue
+                        xsd_node = schema_context.root.maps.attributes.get(name)
+                        if xsd_node is None:
+                            continue
+                    elif xsd_node.value.is_matching(name):
                         if xsd_node.name is None:
                             # node is an XSD attribute wildcard
                             xsd_node = schema_context.root.maps.attributes.get(name)
                             if xsd_node is None:
                                 continue
+                    else:
+                        continue
 
-                        xsd_type = self.add_xsd_type(xsd_node)
+                    xsd_type = self.add_xsd_type(xsd_node)
+                    if xsd_type is not None:
                         value = self.parser.get_atomic_value(xsd_type)
                         yield TypedAttribute(xsd_node, xsd_type, value)
 
+                elif name == XSD_SCHEMA == xsd_node.tag:
+                    # The element is a schema
+                    yield xsd_node
+
                 elif xsd_node.is_matching(name, self.parser.default_namespace):
                     if xsd_node.name is None:
                         # node is an XSD element wildcard
@@ -639,15 +695,14 @@ class XPathToken(Token):
                             continue
 
                     xsd_type = self.add_xsd_type(xsd_node)
-                    value = self.parser.get_atomic_value(xsd_type)
-                    yield TypedElement(xsd_node, xsd_type, value)
+                    if xsd_type is not None:
+                        value = self.parser.get_atomic_value(xsd_type)
+                        yield TypedElement(xsd_node, xsd_type, value)
 
             except AttributeError:
-                # Item is a schema
-                if name == xsd_node.tag == '{%s}schema' % XSD_NAMESPACE:
-                    yield xsd_node
+                pass
 
-    def add_xsd_type(self, item):
+    def add_xsd_type(self, item: Any) -> Optional[XsdTypeProtocol]:
         """
         Adds an XSD type association from an item. The association is
         added using the item's name and type.
@@ -660,23 +715,27 @@ class XPathToken(Token):
             item = item.elem
 
         if not is_schema_node(item):
-            return
+            return None
+
+        name: str = item.name
+        xsd_type: XsdTypeProtocol = item.type
 
         if self.xsd_types is None:
-            self.xsd_types = {item.name: item.type}
+            self.xsd_types = {name: xsd_type}
         else:
-            obj = self.xsd_types.get(item.name)
+            obj = self.xsd_types.get(name)
             if obj is None:
-                self.xsd_types[item.name] = item.type
+                self.xsd_types[name] = xsd_type
             elif not isinstance(obj, list):
-                if obj is not item.type:
-                    self.xsd_types[item.name] = [obj, item.type]
-            elif item.type not in obj:
-                obj.append(item.type)
+                if obj is not xsd_type:
+                    self.xsd_types[name] = [obj, xsd_type]
+            elif xsd_type not in obj:
+                obj.append(xsd_type)
 
-        return item.type
+        return xsd_type
 
-    def get_xsd_type(self, item):
+    def get_xsd_type(self, item: Union[str, PrincipalNodeType]) \
+            -> Optional[XsdTypeProtocol]:
         """
         Returns the XSD type associated with an item. Match by item's name
         and XSD validity. Returns `None` if no XSD type is matching.
@@ -684,18 +743,19 @@ class XPathToken(Token):
         :param item: a string or an AttributeNode or an element.
         """
         if not self.xsd_types or isinstance(self.xsd_types, AbstractSchemaProxy):
-            return
+            return None
         elif isinstance(item, str):
             xsd_type = self.xsd_types.get(item)
         elif isinstance(item, AttributeNode):
             xsd_type = self.xsd_types.get(item.name)
         elif isinstance(item, (TypedAttribute, TypedElement)):
-            return item.xsd_type
+            return cast(XsdTypeProtocol, item.xsd_type)
         else:
             xsd_type = self.xsd_types.get(item.tag)
 
+        x: XsdTypeProtocol
         if not xsd_type:
-            return
+            return None
         elif not isinstance(xsd_type, list):
             return xsd_type
         elif isinstance(item, AttributeNode):
@@ -705,14 +765,14 @@ class XPathToken(Token):
         elif is_etree_element(item):
             for x in xsd_type:
                 if x.is_simple():
-                    if x.is_valid(item.text):
+                    if x.is_valid(item.text):  # type: ignore[union-attr]
                         return x
                 elif x.is_valid(item):
                     return x
 
         return xsd_type[0]
 
-    def get_typed_node(self, item: Union[Element, AttributeNode]):
+    def get_typed_node(self, item: PrincipalNodeType) -> PrincipalNodeType:
         """
         Returns a typed node if the item is matching an XSD type.
 
@@ -756,7 +816,7 @@ class XPathToken(Token):
             atomic_types = xsd11_atomic_types
 
         try:
-            builder = atomic_types[xsd_type.name]
+            builder: Any = atomic_types[xsd_type.name]
         except KeyError:
             pass
         else:
@@ -800,7 +860,7 @@ class XPathToken(Token):
         msg = "Type {!r} does not match sequence type of {!r}"
         raise self.wrong_sequence_type(msg.format(xsd_type, item)) from None
 
-    def cast_to_qname(self, qname):
+    def cast_to_qname(self, qname: str) -> QName:
         """Cast a prefixed qname string to a QName object."""
         try:
             if ':' not in qname:
@@ -813,18 +873,18 @@ class XPathToken(Token):
         except KeyError as err:
             raise self.error('FONS0004', 'no namespace found for prefix {}'.format(err))
 
-    def cast_to_double(self, value):
+    def cast_to_double(self, value: Union[SupportsFloat, str]) -> float:
         """Cast a value to xs:double."""
         try:
             if self.parser.xsd_version == '1.0':
-                return DoubleProxy10(value)
-            return DoubleProxy(value)
+                return cast(float, DoubleProxy10(value))
+            return cast(float, DoubleProxy(value))
         except ValueError as err:
             raise self.error('FORG0001', str(err))  # str or UntypedAtomic
 
     ###
     # XPath data accessors base functions
-    def boolean_value(self, obj):
+    def boolean_value(self, obj: Any) -> bool:
         """
         The effective boolean value, as computed by fn:boolean().
         """
@@ -850,7 +910,7 @@ class XPathToken(Token):
             message = "effective boolean value is not defined for {!r}.".format(type(obj))
             raise self.error('FORG0006', message)
 
-    def data_value(self, obj):
+    def data_value(self, obj: Any) -> Optional[AtomicValueType]:
         """
         The typed value, as computed by fn:data() on each item.
         Returns an instance of UntypedAtomic for untyped data.
@@ -858,30 +918,31 @@ class XPathToken(Token):
         https://www.w3.org/TR/xpath20/#dt-typed-value
         """
         if obj is None:
-            return
+            return None
         elif isinstance(obj, XPathNode):
-            if isinstance(obj, (AttributeNode, TextNode)):
+            if isinstance(obj, TextNode):
                 return UntypedAtomic(obj.value)
-            return obj.value  # a typed node or a NamespaceNode
+            elif isinstance(obj, AttributeNode) and isinstance(obj.value, str):
+                return UntypedAtomic(obj.value)
+            return cast(Optional[AtomicValueType], obj.value)  # a typed node or a NamespaceNode
 
         elif is_schema_node(obj):
             return self.parser.get_atomic_value(obj.type)
 
         elif hasattr(obj, 'tag'):
-            if is_comment_node(obj):
-                return obj.text
-            elif is_processing_instruction_node(obj):
-                return obj.text
+            if is_comment_node(obj) or is_processing_instruction_node(obj):
+                return cast(str, obj.text)
             elif hasattr(obj, 'attrib') and hasattr(obj, 'text'):
                 return UntypedAtomic(''.join(etree_iter_strings(obj)))
-
+            else:
+                return None
         elif is_document_node(obj):
             value = ''.join(etree_iter_strings(obj.getroot()))
             return UntypedAtomic(value)
         else:
-            return obj
+            return cast(AtomicValueType, obj)
 
-    def string_value(self, obj):
+    def string_value(self, obj: Any) -> str:
         """
         The string value, as computed by fn:string().
         """
@@ -895,14 +956,12 @@ class XPathToken(Token):
             elif isinstance(obj, (AttributeNode, TypedAttribute)):
                 return str(obj.value)
             else:
-                return obj.value  # TextNode or NamespaceNode
+                return cast(str, obj.value)  # TextNode or NamespaceNode
         elif is_schema_node(obj):
             return str(self.parser.get_atomic_value(obj.type))
         elif hasattr(obj, 'tag'):
-            if is_comment_node(obj):
-                return obj.text
-            elif is_processing_instruction_node(obj):
-                return obj.text
+            if is_comment_node(obj) or is_processing_instruction_node(obj):
+                return cast(str, obj.text)
             elif hasattr(obj, 'attrib') and hasattr(obj, 'text'):
                 return ''.join(etree_iter_strings(obj))
         elif is_document_node(obj):
@@ -932,7 +991,7 @@ class XPathToken(Token):
 
         return str(obj)
 
-    def number_value(self, obj):
+    def number_value(self, obj: Any) -> float:
         """
         The numeric value, as computed by fn:number() on each item. Returns a float value.
         """
@@ -955,13 +1014,14 @@ class XPathToken(Token):
         return code  # returns an unprefixed code (without prefix the namespace is not checked)
 
     def error(self, code: Union[str, QName],
-              message_or_error: Optional[str] = None) -> ElementPathError:
+              message_or_error: Union[None, str, Exception] = None) -> ElementPathError:
         """
-        Returns an XPath error instance related with a code. An XPath/XQuery/XSLT error code is an
-        alphanumeric token starting with four uppercase letters and ending with four digits.
+        Returns an XPath error instance related with a code. An XPath/XQuery/XSLT
+        error code is an alphanumeric token starting with four uppercase letters
+        and ending with four digits.
 
         :param code: the error code as QName or string.
-        :param message_or_error: an optional custom additional message.
+        :param message_or_error: an optional custom message or an exception.
         """
         namespace: Optional[str]
 
@@ -1010,15 +1070,20 @@ class XPathToken(Token):
         return error_class(message, code=self.error_code(code), token=self)
 
     # Shortcuts for XPath errors, only the wrong_syntax
-    def expected(self, *symbols, message=None, code='XPST0003'):
+    def expected(self, *symbols: str,
+                 message: Optional[str] = None,
+                 code: str = 'XPST0003') -> None:
         if symbols and self.symbol not in symbols:
             raise self.wrong_syntax(message, code)
 
-    def unexpected(self, *symbols, message=None, code='XPST0003'):
+    def unexpected(self, *symbols: str,
+                   message: Optional[str] = None,
+                   code: str = 'XPST0003') -> None:
         if not symbols or self.symbol in symbols:
             raise self.wrong_syntax(message, code)
 
-    def wrong_syntax(self, message=None, code='XPST0003'):
+    def wrong_syntax(self, message: Optional[str] = None,  # type: ignore[override]
+                     code: str = 'XPST0003') -> ElementPathError:
         if self.label == 'function':
             code = 'XPST0017'
 
@@ -1028,55 +1093,35 @@ class XPathToken(Token):
         error = super(XPathToken, self).wrong_syntax(message)
         return self.error(code, str(error))
 
-    def wrong_value(self, message=None):
-        return self.error('FOCA0002', message)
+    def wrong_value(self, message: Optional[str] = None) -> ElementPathValueError:
+        return cast(ElementPathValueError, self.error('FOCA0002', message))
 
-    def wrong_type(self, message=None):
-        return self.error('FORG0006', message)
+    def wrong_type(self, message: Optional[str] = None) -> ElementPathTypeError:
+        return cast(ElementPathTypeError, self.error('FORG0006', message))
 
-    def missing_schema(self, message=None):
-        return self.error('XPST0001', message)
+    def missing_context(self, message: Optional[str] = None) -> MissingContextError:
+        return cast(MissingContextError, self.error('XPDY0002', message))
 
-    def missing_context(self, message=None):
-        return self.error('XPDY0002', message)
+    def wrong_context_type(self, message: Optional[str] = None) -> ElementPathTypeError:
+        return cast(ElementPathTypeError, self.error('XPTY0004', message))
 
-    def wrong_context_type(self, message=None):
-        return self.error('XPTY0004', message)
+    def missing_name(self, message: Optional[str] = None) -> ElementPathNameError:
+        return cast(ElementPathNameError, self.error('XPST0008', message))
 
-    def missing_sequence(self, message=None):
-        return self.error('XPST0005', message)
-
-    def missing_name(self, message=None):
-        return self.error('XPST0008', message)
-
-    def missing_axis(self, message=None):
+    def missing_axis(self, message: Optional[str] = None) \
+            -> Union[ElementPathNameError, ElementPathSyntaxError]:
         if self.parser.compatibility_mode:
-            return self.error('XPST0010', message)
-        return self.error('XPST0003', message)
+            return cast(ElementPathNameError, self.error('XPST0010', message))
+        return cast(ElementPathSyntaxError, self.error('XPST0003', message))
 
-    def wrong_nargs(self, message=None):
-        return self.error('XPST0017', message)
+    def wrong_nargs(self, message: Optional[str] = None) -> ElementPathTypeError:
+        return cast(ElementPathTypeError, self.error('XPST0017', message))
 
-    def wrong_step_result(self, message=None):
-        return self.error('XPTY0018', message)
+    def wrong_sequence_type(self, message: Optional[str] = None) -> ElementPathTypeError:
+        return cast(ElementPathTypeError, self.error('XPDY0050', message))
 
-    def wrong_intermediate_step_result(self, message=None):
-        return self.error('XPTY0019', message)
-
-    def wrong_axis_argument(self, message=None):
-        return self.error('XPTY0020', message)
-
-    def wrong_sequence_type(self, message=None):
-        return self.error('XPDY0050', message)
-
-    def unknown_atomic_type(self, message=None):
-        return self.error('XPST0051', message)
-
-    def wrong_target_type(self, message=None):
-        return self.error('XPST0080', message)
-
-    def unknown_namespace(self, message=None):
-        return self.error('XPST0081', message)
+    def unknown_atomic_type(self, message: Optional[str] = None) -> ElementPathNameError:
+        return cast(ElementPathNameError, self.error('XPST0051', message))
 
 
 class XPathAxis(XPathToken):
@@ -1084,14 +1129,14 @@ class XPathAxis(XPathToken):
     label = 'axis'
     reverse_axis: bool = False
 
-    def nud(self):
+    def nud(self) -> 'XPathAxis':
         self.parser.advance('::')
         self.parser.expected_name(
             '(name)', '*', 'text', 'node', 'document-node',
             'comment', 'processing-instruction', 'attribute',
             'schema-attribute', 'element', 'schema-element'
         )
-        self[:] = self.parser.expression(rbp=self.rbp),
+        self._items[:] = self.parser.expression(rbp=self.rbp),
         return self
 
 
@@ -1137,12 +1182,14 @@ class XPathFunction(XPathToken):
 
     def __call__(self, context: Optional[XPathContext] = None,
                  argument_list: Optional[Union[
-                     XPathToken, List[XPathToken], Tuple[XPathToken, ...]
+                     XPathToken,
+                     List[Union[XPathToken, AtomicValueType]],
+                     Tuple[Union[XPathToken, AtomicValueType], ...]
                  ]] = None) -> Any:
-        args = []
+
+        args: List[Union[Token[XPathTokenType], AtomicValueType]] = []
         if isinstance(argument_list, (list, tuple)):
-            for token in argument_list:
-                args.append(token)
+            args.extend(argument_list)
         elif isinstance(argument_list, XPathToken):
             if argument_list.symbol == '(':
                 args.append(argument_list)
@@ -1160,16 +1207,22 @@ class XPathFunction(XPathToken):
                 if not self.parser.match_sequence_type(value, sequence_type):
                     msg = "invalid type for argument {!r}"
                     raise self.error('XPTY0004', msg.format(variable[0].value))
-                context.variables[variable[0].value] = value
+                varname = cast(str, variable[0].value)
+                context.variables[varname] = value
         elif any(tk.symbol == '?' for tk in self):
             for value, tk in zip(args, filter(lambda x: x.symbol == '?', self)):
-                tk.value = value
+                if isinstance(value, XPathToken):
+                    tk.value = value.evaluate(context)
+                else:
+                    assert not isinstance(value, Token), "An atomic value or None expected"
+                    tk.value = value
         else:
             self.clear()
             for value in args:
                 if isinstance(value, XPathToken):
                     self.append(value)
                 else:
+                    assert not isinstance(value, Token), "An atomic value or None expected"
                     self.append(ValueToken(self.parser, value=value))
 
         result = self.evaluate(context)
@@ -1180,9 +1233,9 @@ class XPathFunction(XPathToken):
         return result
 
     @property
-    def name(self):
+    def name(self) -> Optional[QName]:
         if self.symbol == 'function':
-            return
+            return None
         elif self._name is None:
             if not self.namespace or self.namespace == XPATH_FUNCTIONS_NAMESPACE:
                 self._name = QName(XPATH_FUNCTIONS_NAMESPACE, 'fn:%s' % self.symbol)
@@ -1194,15 +1247,15 @@ class XPathFunction(XPathToken):
         return self._name
 
     @property
-    def arity(self):
+    def arity(self) -> int:
         return self.nargs if isinstance(self.nargs, int) else len(self)
 
-    def nud(self):
+    def nud(self) -> 'XPathFunction':
         code = 'XPST0017' if self.label == 'function' else 'XPST0003'
         self.value = None
         self.parser.advance('(')
         if self.nargs is None:
-            del self[:]
+            del self._items[:]
             if self.parser.next_token.symbol in (')', '(end)'):
                 raise self.error(code, 'at least an argument is required')
             while True:
@@ -1230,7 +1283,7 @@ class XPathFunction(XPathToken):
                 msg = 'Too few arguments: expected at least %s arguments' % min_args
                 raise self.wrong_nargs(msg if min_args > 1 else msg[:-1])
 
-            self[k:] = self.parser.expression(5),
+            self._items[k:] = self.parser.expression(5),
             k += 1
             if k < min_args:
                 if self.parser.next_token.symbol == ')':
@@ -1241,16 +1294,25 @@ class XPathFunction(XPathToken):
         while max_args is None or k < max_args:
             if self.parser.next_token.symbol == ',':
                 self.parser.advance(',')
-                self[k:] = self.parser.expression(5),
+                self._items[k:] = self.parser.expression(5),
             elif k == 0 and self.parser.next_token.symbol != ')':
-                self[k:] = self.parser.expression(5),
+                self._items[k:] = self.parser.expression(5),
             else:
                 break  # pragma: no cover
             k += 1
 
         if self.parser.next_token.symbol == ',':
             msg = 'Too many arguments: expected at most %s arguments' % max_args
-            raise self.error(code, msg if max_args > 1 else msg[:-1])
+            raise self.error(code, msg if max_args != 1 else msg[:-1])
 
         self.parser.advance(')')
         return self
+
+
+class XPathConstructor(XPathFunction):
+    """
+    A token for processing XPath 2.0+ constructors.
+    """
+    @staticmethod
+    def cast(value: Any) -> AtomicValueType:
+        raise NotImplementedError()
